@@ -4,6 +4,7 @@ use strict;
 use warnings;
 
 use Config;
+use Cwd;
 use Getopt::Long;
 use Pod::Usage;
 use File::Spec;
@@ -47,8 +48,8 @@ GetOptions(\%args, qw(
     has-dyncall has-libffi pkgconfig=s
     build=s host=s big-endian jit! enable-jit
     prefix=s bindir=s libdir=s mastdir=s
-    relocatable make-install asan ubsan
-    valgrind telemeh show-autovect
+    relocatable make-install asan ubsan tsan
+    valgrind telemeh show-autovect git-cache-dir=s
     show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
@@ -77,18 +78,9 @@ for my $target (@target_dirs) {
     $config{$target} = $args{$target} if $args{$target};
 }
 
-if (-d '.git') {
-    print dots("Updating submodules");
-    my $msg = qx{git submodule sync --quiet && git submodule --quiet update --init 2>&1};
-    if ($? >> 8 == 0) { print "OK\n" }
-    else {
-        if ($msg =~ /[']([^']+)[']\s+already exists and is not an empty/) {
-            $folder_to_delete = "\n\nERROR: Cannot update submodule because directory exists and is not empty.\n" .
-            ">>> Please delete the following folder and try again:\n$1\n\n";
-        }
-        softfail("git error: $msg")
-    }
-}
+# Download / Update submodules
+my $code = system($^X, 'tools/update-submodules.pl', Cwd::cwd(), @args);
+exit 1 if $code >> 8 != 0;
 
 # fiddle with flags
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
@@ -101,7 +93,7 @@ if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
 }
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
-        has-libatomic_ops asan ubsan valgrind show-vec)) {
+        has-libatomic_ops asan ubsan tsan valgrind show-vec)) {
     $args{$_} = 0 unless defined $args{$_};
 }
 
@@ -389,6 +381,7 @@ $config{ldlibs} = join ' ',
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
 $config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan} && $^O ne 'darwin' && $config{cc} ne 'clang';
 $config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan} and $^O ne 'darwin';
+$config{ldlibs} = ' -ltsan ' . $config{ldlibs} if $args{tsan} and $^O ne 'darwin';
 $config{ldlibs} = $config{ldlibs} . ' -lzstd ' if $config{heapsnapformat} == 3;
 # macro defs
 $config{ccdefflags} = join ' ', map { $config{ccdef} . $_ } @{$config{defs}};
@@ -426,9 +419,10 @@ if ($args{'show-autovect-failed'}) {
     push @cflags, '-Rpass-analysis=loop-vectorize' if 2 <= $args{'show-autovect-failed'} && $config{cc} eq 'clang';
     push @cflags, '-fsave-optimization-record '    if 3 <= $args{'show-autovect-failed'} && $config{cc} eq 'clang';
 }
-push @cflags, '-fno-omit-frame-pointer' if $args{asan} or $args{ubsan};
+push @cflags, '-fno-omit-frame-pointer' if $args{asan} or $args{ubsan} or $args{tsan};
 push @cflags, '-fsanitize=address' if $args{asan};
 push @cflags, '-fsanitize=undefined' if $args{ubsan};
+push @cflags, '-fsanitize=thread' if $args{tsan};
 push @cflags, '-DWSL_BASH_ON_WIN' if wsl_bash_on_win();
 push @cflags, '-DDEBUG_HELPERS' if $args{debug};
 push @cflags, '-DMVM_VALGRIND_SUPPORT' if $args{valgrind};
@@ -450,6 +444,7 @@ if (not $args{static} and $config{prefix} ne '/usr') {
     push @ldflags, $config{ldrpath}             if !$args{relocatable};
 }
 push @ldflags, '-fsanitize=address'  if $args{asan};
+push @ldflags, '-fsanitize=thread'  if $args{tsan};
 push @ldflags, $ENV{LDFLAGS}         if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
@@ -963,8 +958,8 @@ __END__
                    [--static] [--prefix <path>] [--relocatable]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
-                   [--asan] [--ubsan] [--no-jit]
-                   [--telemeh]
+                   [--asan] [--ubsan] [--tsan] [--no-jit]
+                   [--telemeh] [--git-cache-dir <path>]
 
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
@@ -1070,6 +1065,16 @@ A full list of options is displayed if you set C<ASAN_OPTIONS> to C<help=1>.
 
 Build with Undefined Behaviour sanitizer support.
 
+=item --tsan
+
+Build with ThreadSanitizer (TSAN) support. Requires clang and LLVM 3.2 or newer
+alternatively gcc 4.8 or newer. You can use C<TSAN_OPTIONS> to configure TSAN
+at runtime, e.g.
+
+    export TSAN_OPTIONS=report_thread_leaks=0
+
+A full list of options is displayed if you set C<TSAN_OPTIONS> to C<help=1>.
+
 =item --valgrind
 
 Include Valgrind Client Requests for moarvm's own memory allocators.
@@ -1152,4 +1157,13 @@ Disable JIT compiler, which is enabled by default to JIT-compile hot frames.
 
 Build support for the fine-grained internal event logger.
 
+=item --git-cache-dir <path>
+
+Use the given path as a git repository cache.
+For example: --git-cache-dir=/home/user/repo/git_cache
+Each repository ('MoarVM' and its submodules) will use a separate subfolder.
+If the subfolder does not exist, it will be cloned. If it exists the
+contained repository will be updated.
+
+
 =back
