@@ -3,9 +3,9 @@
 /* Gets the HOW (meta-object), which may be lazily deserialized. */
 MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st) {
     MVMObject *HOW = st->HOW;
-    if (!HOW)
-        st->HOW = HOW = MVM_sc_get_object(tc, st->HOW_sc, st->HOW_idx);
-    return HOW;
+    if (!HOW && st->HOW_sc)
+        MVM_ASSIGN_REF(tc, &(st->header), st->HOW, HOW = MVM_sc_get_object(tc, st->HOW_sc, st->HOW_idx));
+    return HOW ? HOW : tc->instance->VMNull;
 }
 
 /* Gets the HOW (meta-object), which may be lazily deserialized, through the
@@ -137,8 +137,8 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
             char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
             char *waste[] = { c_name, NULL };
             MVM_exception_throw_adhoc_free(tc, waste,
-                "Cannot find method '%s': no method cache and no .^find_method",
-                 c_name);
+                "Cannot find method '%s' on '%s': no method cache and no .^find_method",
+                 c_name, MVM_6model_get_debug_name(tc, obj));
         }
         else {
             res->o = tc->instance->VMNull;
