@@ -221,12 +221,11 @@ static void compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *ar
 
 #define introspect_member(member, set_result, result) \
 static void member(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) { \
-    MVMObject *self, *type_obj, *member; \
+    MVMObject *self, *member; \
     MVMArgProcContext arg_ctx; \
     MVM_args_proc_init(tc, &arg_ctx, callsite, args); \
     MVM_args_checkarity(tc, &arg_ctx, 2, 2); \
     self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0); \
-    type_obj = MVM_args_get_required_pos_obj(tc, &arg_ctx, 1); \
     MVM_args_proc_cleanup(tc, &arg_ctx); \
     if (!self || !IS_CONCRETE(self) || REPR(self)->ID != MVM_REPR_ID_KnowHOWREPR) \
         MVM_exception_throw_adhoc(tc, "KnowHOW methods must be called on object instance with REPR KnowHOWREPR"); \
@@ -566,6 +565,7 @@ static void string_consts(MVMThreadContext *tc) {
     string_creator(buf_type, "buf_type");
     string_creator(write, "write");
     string_creator(stdin_fd, "stdin_fd");
+    string_creator(stdin_fd_close, "stdin_fd_close");
     string_creator(stdout_fd, "stdout_fd");
     string_creator(stderr_fd, "stderr_fd");
     string_creator(nativeref, "nativeref");
@@ -576,6 +576,8 @@ static void string_consts(MVMThreadContext *tc) {
     string_creator(ready, "ready");
     string_creator(multidim, "multidim");
     string_creator(entry_point, "entry_point");
+    string_creator(resolve_lib_name, "resolve_lib_name");
+    string_creator(resolve_lib_name_arg, "resolve_lib_name_arg");
     string_creator(kind, "kind");
     string_creator(instrumented, "instrumented");
     string_creator(heap, "heap");
