@@ -187,12 +187,14 @@ static void register_repr(MVMThreadContext *tc, const MVMREPROps *repr, MVMStrin
 
     /* Fill a registry entry. */
     entry = MVM_malloc(sizeof(MVMReprRegistry));
+    MVM_HASH_BIND_FREE(tc, tc->instance->repr_hash, name, entry, {
+        MVM_free(entry);
+    });
     entry->name = name;
     entry->repr = repr;
 
     /* Enter into registry. */
     tc->instance->repr_list[repr->ID] = entry;
-    MVM_HASH_BIND(tc, tc->instance->repr_hash, name, entry);
 
     /* Name and hash key should become a permanent GC root. */
     MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->name, "REPR name");
@@ -291,14 +293,17 @@ static MVMReprRegistry * find_repr_by_name(MVMThreadContext *tc,
         MVMString *name) {
     MVMReprRegistry *entry;
 
+    uv_mutex_lock(&tc->instance->mutex_repr_registry);
     MVM_HASH_GET(tc, tc->instance->repr_hash, name, entry)
 
     if (entry == NULL) {
         char *c_name = MVM_string_ascii_encode_any(tc, name);
         char *waste[] = { c_name, NULL };
+        uv_mutex_unlock(&tc->instance->mutex_repr_registry);
         MVM_exception_throw_adhoc_free(tc, waste, "Lookup by name of unknown REPR: %s",
             c_name);
     }
+    uv_mutex_unlock(&tc->instance->mutex_repr_registry);
 
     return entry;
 }
