@@ -48,28 +48,45 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
     MVM_ASSIGN_REF(tc, &(dest_root->header), dest_body->static_code, src_body->static_code);
 
     dest_body->num_locals = src_body->num_locals;
-    dest_body->num_lexicals = src_body->num_lexicals;
-    {
+    if (dest_body->num_locals) {
         MVMuint16 *local_types = MVM_malloc(sizeof(MVMuint16) * src_body->num_locals);
-        MVMuint16 *lexical_types = MVM_malloc(sizeof(MVMuint16) * src_body->num_lexicals);
         memcpy(local_types, src_body->local_types, sizeof(MVMuint16) * src_body->num_locals);
-        if (src_body->num_lexicals)
-            memcpy(lexical_types, src_body->lexical_types,
-                sizeof(MVMuint16) * src_body->num_lexicals);
         dest_body->local_types = local_types;
-        dest_body->lexical_types = lexical_types;
     }
-    {
-        MVMLexicalRegistry *current, *tmp;
+    else {
+        dest_body->local_types = NULL;
+    }
+
+    MVMuint32 num_lexicals = src_body->num_lexicals;
+    dest_body->num_lexicals = num_lexicals;
+    if (num_lexicals) {
+        int create_hash = src_body->lexical_names != NULL;
+        MVMuint16 *lexical_types = MVM_malloc(sizeof(MVMuint16) * src_body->num_lexicals);
+        MVMLexicalRegistry **lexical_names_list = MVM_malloc(sizeof(MVMLexicalRegistry *) * src_body->num_lexicals);
+        memcpy(lexical_types, src_body->lexical_types,
+               sizeof(MVMuint16) * src_body->num_lexicals);
+        memcpy(lexical_names_list, src_body->lexical_names_list,
+               sizeof(MVMLexicalRegistry *) * src_body->num_lexicals);
+
+        for (MVMuint32 j = 0; j < num_lexicals; j++) {
+            MVMLexicalRegistry *current = lexical_names_list[j];
 
-        /* NOTE: if we really wanted to, we could avoid rehashing... */
-        HASH_ITER_FAST(tc, hash_handle, src_body->lexical_names, current, {
             MVMLexicalRegistry *new_entry = MVM_malloc(sizeof(MVMLexicalRegistry));
+            if (create_hash) {
+                MVM_HASH_BIND_FREE(tc, dest_body->lexical_names, current->key, new_entry, {
+                    MVM_free(new_entry);
+                });
+            }
             /* don't need to clone the string */
             MVM_ASSIGN_REF(tc, &(dest_root->header), new_entry->key, current->key);
             new_entry->value = current->value;
-            MVM_HASH_BIND(tc, dest_body->lexical_names, current->key, new_entry);
-        });
+        }
+        dest_body->lexical_names_list = lexical_names_list;
+        dest_body->lexical_types = lexical_types;
+    }
+    else {
+        dest_body->lexical_names_list = NULL;
+        dest_body->lexical_types = NULL;
     }
 
     /* Static environment needs to be copied, and any objects WB'd. */
@@ -99,10 +116,14 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
         MVM_ASSIGN_REF(tc, &(dest_root->header), dest_body->outer, src_body->outer);
 
     dest_body->num_handlers = src_body->num_handlers;
-    dest_body->handlers     = MVM_malloc(src_body->num_handlers * sizeof(MVMFrameHandler));
-    if (src_body->num_handlers)
+    if (src_body->num_handlers) {
+        dest_body->handlers = MVM_malloc(src_body->num_handlers * sizeof(MVMFrameHandler));
         memcpy(dest_body->handlers, src_body->handlers,
             src_body->num_handlers * sizeof(MVMFrameHandler));
+    }
+    else {
+        dest_body->handlers = NULL;
+    }
     dest_body->instrumentation_level = 0;
     dest_body->num_annotations       = src_body->num_annotations;
     dest_body->annotations_data      = src_body->annotations_data;
@@ -112,7 +133,6 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* Adds held objects to the GC worklist. */
 static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMStaticFrameBody *body = (MVMStaticFrameBody *)data;
-    MVMLexicalRegistry *current;
     MVMStaticFrameDebugLocal *current_debug_local;
 
     /* mvmobjects */
@@ -126,11 +146,15 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     if (!body->fully_deserialized)
         return;
 
-    /* lexical names hash keys */
-    HASH_ITER_FAST(tc, hash_handle, body->lexical_names, current, {
+    /* lexical names */
+    MVMuint32 num_lexicals = body->num_lexicals;
+    MVMLexicalRegistry **lexical_names_list = body->lexical_names_list;
+
+    for (MVMuint32 j = 0; j < num_lexicals; j++) {
+        MVMLexicalRegistry *current = lexical_names_list[j];
         MVM_gc_worklist_add(tc, worklist, &current->hash_handle.key);
         MVM_gc_worklist_add(tc, worklist, &current->key);
-    });
+    }
 
     /* static env */
     if (body->static_env) {
@@ -173,7 +197,8 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVM_free(body->local_types);
     MVM_free(body->lexical_types);
     MVM_free(body->lexical_names_list);
-    MVM_HASH_DESTROY(tc, hash_handle, MVMLexicalRegistry, body->lexical_names);
+    if (body->lexical_names)
+        MVM_HASH_DESTROY(tc, hash_handle, MVMLexicalRegistry, body->lexical_names);
 }
 
 static const MVMStorageSpec storage_spec = {
@@ -211,7 +236,7 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
         size += sizeof(MVMLexicalRegistry *) * body->num_lexicals;
 
-        size += sizeof(MVMLexicalRegistry) * HASH_CNT(hash_handle, body->lexical_names);
+        size += sizeof(MVMLexicalRegistry) * body->num_lexicals;
 
         size += sizeof(MVMFrameHandler) * body->num_handlers;
 
@@ -238,7 +263,6 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
 static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTable *st, void *data) {
     MVMStaticFrameBody *body = (MVMStaticFrameBody *)data;
-    MVMLexicalRegistry *current;
 
     static MVMuint64 cache_1 = 0;
     static MVMuint64 cache_2 = 0;
@@ -264,18 +288,21 @@ static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTa
     if (!body->fully_deserialized)
         return;
 
-    /* lexical names hash keys */
-    HASH_ITER_FAST(tc, hash_handle, body->lexical_names, current, {
+    /* lexical names */
+    MVMuint32 num_lexicals = body->num_lexicals;
+    MVMLexicalRegistry **lexical_names_list = body->lexical_names_list;
+
+    for (MVMuint32 j = 0; j < num_lexicals; j++) {
+        MVMLexicalRegistry *current = lexical_names_list[j];
         MVM_profile_heap_add_collectable_rel_const_cstr_cached(tc, ss,
             (MVMCollectable *)current->key, "Lexical name", &nonstatic_cache_1);
-    });
+    }
 
     /* static env */
     if (body->static_env) {
         MVMuint16 *type_map = body->lexical_types;
         MVMuint16  count    = body->num_lexicals;
         MVMuint16  i;
-        MVMuint64 cache_1;
         for (i = 0; i < count; i++)
             if (type_map[i] == MVM_reg_str || type_map[i] == MVM_reg_obj)
                 MVM_profile_heap_add_collectable_rel_const_cstr_cached(tc, ss,
@@ -326,7 +353,7 @@ static const MVMREPROps MVMStaticFrame_this_repr = {
 char * MVM_staticframe_file_location(MVMThreadContext *tc, MVMStaticFrame *sf) {
     MVMBytecodeAnnotation *ann = MVM_bytecode_resolve_annotation(tc, &sf->body, 0);
     MVMCompUnit *cu = sf->body.cu;
-    MVMint32           str_idx = ann ? ann->filename_string_heap_index : 0;
+    MVMuint32          str_idx = ann ? ann->filename_string_heap_index : 0;
     MVMint32           line_nr = ann ? ann->line_number : 1;
     MVMString        *filename = cu->body.filename;
     char        *filename_utf8 = "<unknown>";
@@ -340,3 +367,25 @@ char * MVM_staticframe_file_location(MVMThreadContext *tc, MVMStaticFrame *sf) {
         MVM_free(filename_utf8);
     return result;
 }
+
+/* We could change this code (and bytecode.c) to lazily only build the lookup
+ * hash on the first lookup. I don't have a feel for how often no lookups are
+ * made, and hence whether the added complexity would be much of a saving. */
+MVMLexicalRegistry *MVM_get_lexical_by_name(MVMThreadContext *tc, MVMStaticFrame *sf, MVMString *name) {
+    MVMLexicalRegistry *entry;
+    /* deserialize_frames in bytecode.c doesn't create the lookup hash if there
+     * are only a small number of lexicals in this frame. */
+    if (sf->body.lexical_names) {
+        MVM_HASH_GET(tc, sf->body.lexical_names, name, entry);
+        return entry;
+    }
+
+    MVMLexicalRegistry **lexical_names_list = sf->body.lexical_names_list;
+    MVMuint32 num_lexicals = sf->body.num_lexicals;
+    for (MVMuint32 j = 0; j < num_lexicals; j++) {
+        entry = lexical_names_list[j];
+        if (MVM_string_equal(tc, name, entry->key))
+            return entry;
+    }
+    return NULL;
+}
