@@ -28,8 +28,16 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     MVM_spesh_stats_gc_mark(tc, body->spesh_stats, worklist);
     MVM_spesh_arg_guard_gc_mark(tc, body->spesh_arg_guard, worklist);
     if (body->num_spesh_candidates) {
-        MVMint32 i, j;
+        MVMuint32 i, j;
         for (i = 0; i < body->num_spesh_candidates; i++) {
+            if (body->spesh_candidates[i]->type_tuple) {
+                for (j = 0; j < body->spesh_candidates[i]->cs->flag_count; j++) {
+                    MVM_gc_worklist_add(tc, worklist,
+                        &body->spesh_candidates[i]->type_tuple[j].type);
+                    MVM_gc_worklist_add(tc, worklist,
+                        &body->spesh_candidates[i]->type_tuple[j].decont_type);
+                }
+            }
             for (j = 0; j < body->spesh_candidates[i]->num_spesh_slots; j++)
                 MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i]->spesh_slots[j]);
             for (j = 0; j < body->spesh_candidates[i]->num_inlines; j++)
@@ -42,7 +50,7 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
 /* Called by the VM in order to free memory associated with this object. */
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMStaticFrameSpesh *sfs = (MVMStaticFrameSpesh *)obj;
-    MVMint32 i;
+    MVMuint32 i;
     MVM_spesh_stats_destroy(tc, sfs->body.spesh_stats);
     MVM_free(sfs->body.spesh_stats);
     MVM_spesh_arg_guard_destroy(tc, sfs->body.spesh_arg_guard, 0);
@@ -128,7 +136,7 @@ static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTa
     MVM_spesh_arg_guard_gc_describe(tc, ss, body->spesh_arg_guard);
 
     if (body->num_spesh_candidates) {
-        MVMint32 i, j;
+        MVMuint32 i, j;
         for (i = 0; i < body->num_spesh_candidates; i++) {
             for (j = 0; j < body->spesh_candidates[i]->num_spesh_slots; j++)
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
