@@ -21,13 +21,20 @@ MVM_STATIC_INLINE void set_obj_at_offset(MVMThreadContext *tc, MVMObject *root,
     MVM_ASSIGN_REF(tc, &(root->header), *((MVMObject **)location), value);
 }
 
+/* Helper for aligning sizes */
+MVM_STATIC_INLINE void align_to(MVMuint64 *size, MVMuint32 align) {
+    if (*size % align) {
+        *size += align - *size % align;
+    }
+}
+
 /* Helper for finding a slot number. */
 static MVMint64 try_get_slot(MVMThreadContext *tc, MVMP6opaqueREPRData *repr_data, MVMObject *class_key, MVMString *name) {
     if (repr_data->name_to_index_mapping) {
         MVMP6opaqueNameMap *cur_map_entry = repr_data->name_to_index_mapping;
         while (cur_map_entry->class_key != NULL) {
             if (cur_map_entry->class_key == class_key) {
-                MVMint16 i;
+                MVMuint16 i;
                 for (i = 0; i < cur_map_entry->num_attrs; i++) {
                     if (MVM_string_equal(tc, cur_map_entry->names[i], name)) {
                         return cur_map_entry->slots[i];
@@ -162,7 +169,7 @@ static void gc_mark_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist
     if (repr_data->name_to_index_mapping) {
         MVMP6opaqueNameMap *cur_map_entry = repr_data->name_to_index_mapping;
         while (cur_map_entry->class_key != NULL) {
-            MVMint16 i;
+            MVMuint16 i;
             for (i = 0; i < cur_map_entry->num_attrs; i++) {
                 MVM_gc_worklist_add(tc, worklist, &cur_map_entry->names[i]);
             }
@@ -202,6 +209,30 @@ static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
     MVM_free(st->REPR_data);
 }
 
+/* Frees the representation data.*/
+static void free_repr_data(MVMP6opaqueREPRData *repr_data) {
+    if (repr_data->name_to_index_mapping) {
+        MVMP6opaqueNameMap *cur_map_entry = repr_data->name_to_index_mapping;
+        while (cur_map_entry->class_key != NULL) {
+            MVM_free(cur_map_entry->names);
+            MVM_free(cur_map_entry->slots);
+            cur_map_entry++;
+        }
+        MVM_free(repr_data->name_to_index_mapping);
+    }
+
+    MVM_free(repr_data->attribute_offsets);
+    MVM_free(repr_data->flattened_stables);
+    MVM_free(repr_data->auto_viv_values);
+    MVM_free(repr_data->unbox_slots);
+    MVM_free(repr_data->gc_obj_mark_offsets);
+    MVM_free(repr_data->initialize_slots);
+    MVM_free(repr_data->gc_mark_slots);
+    MVM_free(repr_data->gc_cleanup_slots);
+
+    MVM_free(repr_data);
+}
+
 /* Helper for complaining about attribute access errors. */
 MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, MVMSTable *target_type) MVM_NO_RETURN_ATTRIBUTE;
 MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, MVMSTable *target_type) {
@@ -640,9 +671,10 @@ static MVMuint16 * allocate_unbox_slots() {
 }
 static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
     MVMint64   mro_pos, mro_count, num_parents, total_attrs, num_attrs,
-               cur_slot, cur_type, cur_alloc_addr, cur_obj_attr,
+               cur_slot, cur_type, cur_obj_attr,
                cur_init_slot, cur_mark_slot, cur_cleanup_slot,
                unboxed_type, i;
+    MVMuint64  cur_alloc_addr;
     MVMObject *info;
     MVMP6opaqueREPRData *repr_data;
 
@@ -659,14 +691,14 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
     if (st->REPR_data)
         MVM_exception_throw_adhoc(tc, "Type %s is already composed", MVM_6model_get_stable_debug_name(tc, st));
 
-    /* Allocate the representation data. */
-    repr_data = (MVMP6opaqueREPRData *)MVM_calloc(1, sizeof(MVMP6opaqueREPRData));
-
     /* Find attribute information. */
     info = MVM_repr_at_key_o(tc, info_hash, str_attribute);
     if (MVM_is_null(tc, info))
         MVM_exception_throw_adhoc(tc, "P6opaque: missing attribute protocol in compose of %s", MVM_6model_get_stable_debug_name(tc, st));
 
+    /* Allocate the representation data. */
+    repr_data = (MVMP6opaqueREPRData *)MVM_calloc(1, sizeof(MVMP6opaqueREPRData));
+
     /* In this first pass, we'll loop over the MRO entries, looking for
      * if there is any multiple inheritance and counting the number of
      * attributes. */
@@ -753,8 +785,14 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             MVMuint32 align;
 
             /* Ensure we have a name. */
-            if (MVM_is_null(tc, name_obj))
+            if (MVM_is_null(tc, name_obj)) {
+                if (num_attrs) {
+                    MVM_free_null(name_map->names);
+                    MVM_free_null(name_map->slots);
+                }
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc, "P6opaque: %s missing attribute name for attribute %"PRId64, MVM_6model_get_stable_debug_name(tc, st), i);
+            }
 
             if (REPR(name_obj)->ID == MVM_REPR_ID_MVMString) {
                 MVM_ASSIGN_REF(tc, &(st->header), name_map->names[i], (MVMString *)name_obj);
@@ -800,21 +838,39 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                         /* If it boxes a primitive, note that. */
                         switch (unboxed_type) {
                             case MVM_STORAGE_SPEC_BP_INT:
-                                if (repr_data->unbox_int_slot >= 0)
+                                if (repr_data->unbox_int_slot >= 0) {
+                                    if (num_attrs) {
+                                        MVM_free_null(name_map->names);
+                                        MVM_free_null(name_map->slots);
+                                    }
+                                    free_repr_data(repr_data);
                                     MVM_exception_throw_adhoc(tc,
                                         "While composing %s: Duplicate box_target for native int: attributes %d and %"PRId64, MVM_6model_get_stable_debug_name(tc, st), repr_data->unbox_int_slot, i);
+                                }
                                 repr_data->unbox_int_slot = cur_slot;
                                 break;
                             case MVM_STORAGE_SPEC_BP_NUM:
-                                if (repr_data->unbox_num_slot >= 0)
+                                if (repr_data->unbox_num_slot >= 0) {
+                                    if (num_attrs) {
+                                        MVM_free_null(name_map->names);
+                                        MVM_free_null(name_map->slots);
+                                    }
+                                    free_repr_data(repr_data);
                                     MVM_exception_throw_adhoc(tc,
                                         "While composing %s: Duplicate box_target for native num: attributes %d and %"PRId64, MVM_6model_get_stable_debug_name(tc, st), repr_data->unbox_num_slot, i);
+                                }
                                 repr_data->unbox_num_slot = cur_slot;
                                 break;
                             case MVM_STORAGE_SPEC_BP_STR:
-                                if (repr_data->unbox_str_slot >= 0)
+                                if (repr_data->unbox_str_slot >= 0) {
+                                    if (num_attrs) {
+                                        MVM_free_null(name_map->names);
+                                        MVM_free_null(name_map->slots);
+                                    }
+                                    free_repr_data(repr_data);
                                     MVM_exception_throw_adhoc(tc,
                                         "While composing %s: Duplicate box_target for native str: attributes %d and %"PRId64, MVM_6model_get_stable_debug_name(tc, st), repr_data->unbox_str_slot, i);
+                                }
                                 repr_data->unbox_str_slot = cur_slot;
                                 break;
                             default:
@@ -833,9 +889,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             /* C structure needs careful alignment. If cur_alloc_addr is not
              * aligned to align bytes (cur_alloc_addr % align), make sure it is
              * before we add the next element. */
-            if (cur_alloc_addr % align) {
-                cur_alloc_addr += align - cur_alloc_addr % align;
-            }
+            align_to(&cur_alloc_addr, align);
 
             /* Attribute will live at the current position in the object. */
             repr_data->attribute_offsets[cur_slot] = cur_alloc_addr;
@@ -851,27 +905,55 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
 
             /* Is it a positional or associative delegate? */
             if (MVM_repr_exists_key(tc, attr_info, str_pos_del)) {
-                if (repr_data->pos_del_slot != -1)
+                if (repr_data->pos_del_slot != -1) {
+                    if (num_attrs) {
+                        MVM_free_null(name_map->names);
+                        MVM_free_null(name_map->slots);
+                    }
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "While composing %s: Duplicate positional delegate attributes: %d and %"PRId64"", MVM_6model_get_stable_debug_name(tc, st), repr_data->pos_del_slot, cur_slot);
+                }
                 if (unboxed_type == MVM_STORAGE_SPEC_BP_NONE)
                     repr_data->pos_del_slot = cur_slot;
-                else
+                else {
+                    if (num_attrs) {
+                        MVM_free_null(name_map->names);
+                        MVM_free_null(name_map->slots);
+                    }
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "While composing %s: Positional delegate attribute must be a reference type", MVM_6model_get_stable_debug_name(tc, st));
+                }
             }
             if (MVM_repr_exists_key(tc, attr_info, str_ass_del)) {
-                if (repr_data->ass_del_slot != -1)
+                if (repr_data->ass_del_slot != -1) {
+                    if (num_attrs) {
+                        MVM_free_null(name_map->names);
+                        MVM_free_null(name_map->slots);
+                    }
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "While composing %s: Duplicate associative delegate attributes: %d and %"PRId64, MVM_6model_get_stable_debug_name(tc, st), repr_data->pos_del_slot, cur_slot);
+                }
                 if (unboxed_type == MVM_STORAGE_SPEC_BP_NONE)
                     repr_data->ass_del_slot = cur_slot;
-                else
+                else {
+                    if (num_attrs) {
+                        MVM_free_null(name_map->names);
+                        MVM_free_null(name_map->slots);
+                    }
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "While composing %s: Associative delegate attribute must be a reference type", MVM_6model_get_stable_debug_name(tc, st));
+                }
             }
 
-            /* Add the required space for this type. */
+            /* Add the required space for this type.
+             *
+             * Note: This is the only place in this function where cur_alloc_addr can
+             * become unaligned.
+             */
             cur_alloc_addr += bits / 8;
 
             /* Increment slot count. */
@@ -882,8 +964,11 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
         cur_type++;
     }
 
+    align_to(&cur_alloc_addr, ALIGNOF(void *));
+
     /* Add allocated amount for body to have total object size. */
     st->size = sizeof(MVMP6opaque) + (cur_alloc_addr - sizeof(MVMP6opaqueBody));
+    MVM_ASSERT_ALIGNED(st->size, ALIGNOF(void *));
 
     /* Add sentinels/counts. */
     repr_data->gc_obj_mark_offsets_count = cur_obj_attr;
@@ -903,7 +988,7 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
     /* To calculate size, we need number of attributes and to know about
      * anything flattend in. */
     MVMint64  num_attributes = MVM_serialization_read_int(tc, reader);
-    MVMuint32 cur_offset = sizeof(MVMP6opaque);
+    MVMuint64 cur_offset = sizeof(MVMP6opaque);
     MVMint64  i;
     for (i = 0; i < num_attributes; i++) {
         if (MVM_serialization_read_int(tc, reader)) {
@@ -911,9 +996,7 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
             const MVMStorageSpec *ss = st->REPR->get_storage_spec(tc, st);
             if (ss->inlineable) {
                 /* TODO: Review if/when we get sub-byte things. */
-                if (cur_offset % ss->align) {
-                    cur_offset += ss->align - cur_offset % ss->align;
-                }
+                align_to(&cur_offset, (MVMuint32)ss->align);
                 cur_offset += ss->bits / 8;
             }
             else
@@ -924,7 +1007,9 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
         }
     }
 
+    align_to(&cur_offset, ALIGNOF(void *));
     st->size = cur_offset;
+    MVM_ASSERT_ALIGNED(st->size, ALIGNOF(void *));
 }
 
 /* Serializes the REPR data. */
@@ -1000,10 +1085,11 @@ static void serialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializ
 
 /* Deserializes representation data. */
 static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
-    MVMuint16 i, j, num_classes, cur_offset;
+    MVMuint16 i, j, num_classes;
+    MVMuint64 cur_offset;
     MVMint16 cur_initialize_slot, cur_gc_mark_slot, cur_gc_cleanup_slot;
 
-    MVMP6opaqueREPRData *repr_data = MVM_malloc(sizeof(MVMP6opaqueREPRData));
+    MVMP6opaqueREPRData *repr_data = MVM_calloc(1, sizeof(MVMP6opaqueREPRData));
 
     repr_data->num_attributes = (MVMuint16)MVM_serialization_read_int(tc, reader);
 
@@ -1035,12 +1121,16 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
         for (i = 0; i < repr_data->num_attributes; i++) {
             MVMuint16 repr_id = MVM_serialization_read_int(tc, reader);
             MVMuint16 slot = MVM_serialization_read_int(tc, reader);
-            if (slot > repr_data->num_attributes)
+            if (slot > repr_data->num_attributes) {
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc, "Serialization error: P6opaque's unbox slot out of range (slot %d > %d attributes).", slot, repr_data->num_attributes);
+            }
             if (repr_id < MVM_REPR_MAX_COUNT)
                 repr_data->unbox_slots[repr_id] = slot;
-            else
+            else {
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc, "Serialization error: P6opaque's unbox slot repr id out of range.");
+            }
         }
     } else {
         repr_data->unbox_slots = NULL;
@@ -1110,12 +1200,11 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
                 repr_data->gc_cleanup_slots[cur_gc_cleanup_slot++] = i;
 
             if (spec->align == 0) {
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc, "Serialization error: Storage Spec of P6opaque must not have align set to 0.");
             }
 
-            if (cur_offset % spec->align) {
-                cur_offset += spec->align - cur_offset % spec->align;
-            }
+            align_to(&cur_offset, spec->align);
 
             repr_data->attribute_offsets[i] = cur_offset;
 
@@ -1941,7 +2030,7 @@ MVMuint32 MVM_p6opaque_offset_to_attr_idx(MVMThreadContext *tc, MVMObject *type,
 
 #ifdef DEBUG_HELPERS
 /* This is meant to be called in a debugging session and not used anywhere else.
- * Plese don't delete. */
+ * Please don't delete. */
 static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)STABLE(obj)->REPR_data;
     MVMP6opaqueBody *data = MVM_p6opaque_real_data(tc, OBJECT_BODY(obj));
@@ -1956,11 +2045,10 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
         }
         fprintf(stderr, "%s.new(", MVM_6model_get_debug_name(tc, obj));
         if (name_to_index_mapping != NULL) {
-            MVMint16 i;
             MVMP6opaqueNameMap *cur_map_entry = name_to_index_mapping;
 
             while (cur_map_entry->class_key != NULL) {
-                MVMint16 i;
+                MVMuint32 i;
                 MVMint64 slot;
                 if (cur_map_entry->num_attrs > 0) {
                     fprintf(stderr, "#`(from %s) ", MVM_6model_get_stable_debug_name(tc, cur_map_entry->class_key->st));
