@@ -17,9 +17,12 @@ MVMObject * MVM_sc_create(MVMThreadContext *tc, MVMString *handle) {
             uv_mutex_lock(&tc->instance->mutex_sc_registry);
             MVM_HASH_GET(tc, tc->instance->sc_weakhash, handle, scb);
             if (!scb) {
-                sc->body = scb = MVM_calloc(1, sizeof(MVMSerializationContextBody));
+                scb = MVM_calloc(1, sizeof(MVMSerializationContextBody));
                 MVM_ASSIGN_REF(tc, &(sc->common.header), scb->handle, handle);
-                MVM_HASH_BIND(tc, tc->instance->sc_weakhash, handle, scb);
+                MVM_HASH_BIND_FREE(tc, tc->instance->sc_weakhash, handle, scb, {
+                    MVM_free(scb);
+                });
+                sc->body = scb;
                 /* Calling repr_init will allocate, BUT if it does so, and we
                  * get unlucky, the GC will try to acquire mutex_sc_registry.
                  * This deadlocks. Thus, we force allocation in gen2, which
@@ -101,7 +104,7 @@ MVMint64 MVM_sc_find_object_idx(MVMThreadContext *tc, MVMSerializationContext *s
     MVMObject **roots;
     MVMint64    i, count;
     MVMuint32   cached = MVM_sc_get_idx_in_sc(&obj->header);
-    if (cached != ~0 && MVM_sc_get_collectable_sc(tc, &obj->header) == sc)
+    if (cached != ~(unsigned)0 && MVM_sc_get_collectable_sc(tc, &obj->header) == sc)
         return cached;
     roots = sc->body->root_objects;
     count = sc->body->num_objects;
@@ -124,7 +127,7 @@ MVMint64 MVM_sc_find_object_idx_jit(MVMThreadContext *tc, MVMObject *sc, MVMObje
 MVMint64 MVM_sc_find_stable_idx(MVMThreadContext *tc, MVMSerializationContext *sc, MVMSTable *st) {
     MVMuint64 i;
     MVMuint32 cached = MVM_sc_get_idx_in_sc(&st->header);
-    if (cached != ~0 && MVM_sc_get_collectable_sc(tc, &st->header) == sc)
+    if (cached != ~(unsigned)0 && MVM_sc_get_collectable_sc(tc, &st->header) == sc)
         return cached;
     for (i = 0; i < sc->body->num_stables; i++)
         if (sc->body->root_stables[i] == st)
@@ -138,7 +141,7 @@ MVMint64 MVM_sc_find_code_idx(MVMThreadContext *tc, MVMSerializationContext *sc,
     MVMObject *roots;
     MVMint64   i, count;
     MVMuint32 cached = MVM_sc_get_idx_in_sc(&obj->header);
-    if (cached != ~0 && MVM_sc_get_collectable_sc(tc, &obj->header) == sc)
+    if (cached != ~(unsigned)0 && MVM_sc_get_collectable_sc(tc, &obj->header) == sc)
         return cached;
     roots = sc->body->root_codes;
     count = MVM_repr_elems(tc, roots);
@@ -236,15 +239,15 @@ void MVM_sc_set_object(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint
 void MVM_sc_set_object_no_update(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx, MVMObject *obj) {
     if (idx < 0)
         MVM_exception_throw_adhoc(tc, "Invalid (negative) object root index %"PRId64"", idx);
-    if (idx < sc->body->num_objects) {
+    if ((MVMuint64)idx < sc->body->num_objects) {
         /* Just updating an existing one. */
         MVM_ASSIGN_REF(tc, &(sc->common.header), sc->body->root_objects[idx], obj);
     }
     else {
-        if (idx >= sc->body->alloc_objects) {
+        if ((MVMuint64)idx >= sc->body->alloc_objects) {
             MVMint64 orig_size = sc->body->alloc_objects;
             sc->body->alloc_objects *= 2;
-            if (sc->body->alloc_objects < idx + 1)
+            if (sc->body->alloc_objects < (MVMuint64)idx + 1)
                 sc->body->alloc_objects = idx + 1;
             sc->body->root_objects = MVM_recalloc(sc->body->root_objects,
                 orig_size * sizeof(MVMObject *), sc->body->alloc_objects * sizeof(MVMObject *));
@@ -256,7 +259,7 @@ void MVM_sc_set_object_no_update(MVMThreadContext *tc, MVMSerializationContext *
 
 /* Given an SC and an index, fetch the STable stored there. */
 MVMSTable * MVM_sc_get_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
-    if (MVM_LIKELY(idx >= 0 && idx < sc->body->num_stables)) {
+    if (MVM_LIKELY(idx >= 0 && (MVMuint64)idx < sc->body->num_stables)) {
         MVMSTable *got = sc->body->root_stables[idx];
         return got && !sc_working(sc) ? got : MVM_serialization_demand_stable(tc, sc, idx);
     }
@@ -272,7 +275,7 @@ MVMSTable * MVM_sc_get_stable(MVMThreadContext *tc, MVMSerializationContext *sc,
 /* Given an SC and an index, fetch the STable stored there, or return NULL if there
  * is none. Does not cause lazy deserialization. */
 MVMSTable * MVM_sc_try_get_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
-    if (idx >= 0 && idx < sc->body->num_stables)
+    if (idx >= 0 && (MVMuint64)idx < sc->body->num_stables)
         return sc->body->root_stables[idx];
     else
         return NULL;
@@ -284,15 +287,15 @@ void MVM_sc_set_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint
     if (MVM_UNLIKELY(idx < 0))
         MVM_exception_throw_adhoc(tc,
             "Invalid (negative) STable index %"PRId64, idx);
-    if (idx < sc->body->num_stables) {
+    if ((MVMuint64)idx < sc->body->num_stables) {
         /* Just updating an existing one. */
         MVM_ASSIGN_REF(tc, &(sc->common.header), sc->body->root_stables[idx], st);
     }
     else {
-        if (idx >= sc->body->alloc_stables) {
+        if ((MVMuint64)idx >= sc->body->alloc_stables) {
             MVMint64 orig_size = sc->body->alloc_stables;
             sc->body->alloc_stables += 32;
-            if (sc->body->alloc_stables < idx + 1)
+            if (sc->body->alloc_stables < (MVMuint64)idx + 1)
                 sc->body->alloc_stables = idx + 1;
             sc->body->root_stables = MVM_realloc(sc->body->root_stables,
                 sc->body->alloc_stables * sizeof(MVMSTable *));
@@ -308,7 +311,7 @@ void MVM_sc_set_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint
 /* Given an SC and an STable, pushes the STable to the end of the root list. */
 void MVM_sc_push_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMSTable *st) {
     MVMint64 idx = sc->body->num_stables;
-    if (idx == sc->body->alloc_stables) {
+    if ((MVMuint64)idx == sc->body->alloc_stables) {
         sc->body->alloc_stables += 16;
         sc->body->root_stables = MVM_realloc(sc->body->root_stables,
             sc->body->alloc_stables * sizeof(MVMSTable *));
@@ -321,7 +324,7 @@ void MVM_sc_push_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMST
 MVMObject * MVM_sc_get_code(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
     MVMObject *roots = sc->body->root_codes;
     MVMuint64   count = MVM_repr_elems(tc, roots);
-    if (idx < count) {
+    if ((MVMuint64)idx < count) {
         MVMObject *found = MVM_repr_at_pos_o(tc, roots, idx);
         return MVM_is_null(tc, found) || sc_working(sc)
             ? MVM_serialization_demand_code(tc, sc, idx)
@@ -399,7 +402,7 @@ void MVM_sc_disclaim(MVMThreadContext *tc, MVMSerializationContext *sc) {
 /* SC repossession barrier. */
 void MVM_SC_WB_OBJ(MVMThreadContext *tc, MVMObject *obj) {
     assert(!(obj->header.flags & MVM_CF_FORWARDER_VALID));
-    assert(MVM_sc_get_idx_of_sc(&obj->header) != ~0);
+    assert(MVM_sc_get_idx_of_sc(&obj->header) != (MVMuint32)~0);
     if (MVM_sc_get_idx_of_sc(&obj->header) > 0 && !(obj->st->mode_flags & MVM_NEVER_REPOSSESS_TYPE))
         MVM_sc_wb_hit_obj(tc, obj);
 }
