@@ -852,6 +852,8 @@ static void flatten_args(MVMThreadContext *tc, MVMArgProcContext *ctx) {
             MVMStorageSpec  lss   = REPR(list)->pos_funcs.get_elem_storage_spec(tc, STABLE(list));
 
             if ((MVMint64)new_arg_pos + count > 0xFFFF) {
+                MVM_free(new_arg_flags);
+                MVM_free(new_args);
                 MVM_exception_throw_adhoc(tc, "Too many arguments (%"PRId64") in flattening array, only %"PRId32" allowed.", (MVMint64)new_arg_pos + count, 0xFFFF);
             }
 
@@ -929,6 +931,8 @@ static void flatten_args(MVMThreadContext *tc, MVMArgProcContext *ctx) {
                 });
             }
             else if (arg_info.arg.o) {
+                MVM_free(new_arg_flags);
+                MVM_free(new_args);
                 MVM_exception_throw_adhoc(tc, "flattening of other hash reprs NYI.");
             }
         }
