@@ -184,7 +184,6 @@ static char * u64toa_naive_worker(uint64_t value, char* buffer) {
         *buffer++ = *--p;
     } while (p != temp);
 
-    *buffer = '\0';
     return buffer;
 }
 static size_t i64toa_naive(int64_t value, char* buffer) {
@@ -201,7 +200,7 @@ static size_t u64toa_naive(uint64_t value, char* buffer) {
     return u64toa_naive_worker(value, buffer) - buffer;
 }
 MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
-    char buffer[64];
+    char buffer[20];
     int len;
     /* See if we can hit the cache. */
     int cache = 0 <= i && i < MVM_INT_TO_STR_CACHE_SIZE;
@@ -227,7 +226,7 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
 }
 
 MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
-    char buffer[64];
+    char buffer[20];
     int len;
     /* See if we can hit the cache. */
     int cache = i < MVM_INT_TO_STR_CACHE_SIZE;
@@ -326,7 +325,7 @@ void MVM_coerce_smart_stringify(MVMThreadContext *tc, MVMObject *obj, MVMRegiste
 }
 
 MVMint64 MVM_coerce_s_i(MVMThreadContext *tc, MVMString *s) {
-    char     *enc = MVM_string_ascii_encode(tc, s, NULL, 0);
+    char     *enc = MVM_string_ascii_encode_any(tc, s);
     MVMint64  i   = strtoll(enc, NULL, 10);
     MVM_free(enc);
     return i;
@@ -470,8 +469,8 @@ MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMi
     }
 
     ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
-    if ((flag & 0x02) && (ch == '+' || ch == '-')) {
-        neg = (ch == '-');
+    if ((flag & 0x02) && (ch == '+' || ch == '-' || ch == 0x2212)) {  /* MINUS SIGN */
+        neg = (ch == '-' || ch == 0x2212);
         offset++;
         ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
     }
