@@ -25,8 +25,11 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
     MVMCompUnit        *cu                = static_frame_body->cu;
 
     /* Ensure the frame is fully deserialized. */
-    if (!static_frame_body->fully_deserialized)
-        MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
+    if (!static_frame_body->fully_deserialized) {
+        MVMROOT(tc, static_frame, {
+            MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
+        });
+    }
 
     /* If we never invoked this compilation unit before, and we have spesh
      * enabled, we might either have no spesh log or a nearly full one. This
@@ -39,11 +42,9 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
 
     /* Take compilation unit lock, to make sure we don't race to do the
      * frame preparation/verification work. */
-    MVMROOT(tc, cu, {
     MVMROOT(tc, static_frame, {
         MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
     });
-    });
 
     if (static_frame->body.instrumentation_level == 0) {
         /* Work size is number of locals/registers plus size of the maximum
@@ -52,7 +53,9 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
             (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
 
         /* Validate the bytecode. */
-        MVM_validate_static_frame(tc, static_frame);
+        MVMROOT(tc, static_frame, {
+            MVM_validate_static_frame(tc, static_frame);
+        });
 
         /* Compute work area initial state that we can memcpy into place each
          * time. */
@@ -78,7 +81,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
          * also because then it can be ssigned into the gen2 static frame
          * without causing it to become an inter-gen root. */
         MVM_gc_allocate_gen2_default_set(tc);
-        MVM_ASSIGN_REF(tc, &(static_frame->common.header), static_frame_body->spesh,
+        MVM_ASSIGN_REF(tc, &(static_frame->common.header), static_frame_body->spesh, /* no GC error */
             MVM_repr_alloc_init(tc, tc->instance->StaticFrameSpesh));
         MVM_gc_allocate_gen2_default_clear(tc);
 
@@ -96,7 +99,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
  * profiling. */
 static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *static_frame) {
     MVMCompUnit *cu = static_frame->body.cu;
-    MVMROOT2(tc, static_frame, cu, {
+    MVMROOT(tc, static_frame, {
         /* Obtain mutex, so we don't end up with instrumentation races. */
         MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
 
@@ -181,8 +184,8 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
                 MVMuint16 num_lexicals = static_frame->body.num_lexicals;
                 for (i = 0; i < num_lexicals; i++) {
                     if (!static_frame->body.static_env[i].o && static_frame->body.static_env_flags[i] == 1) {
-                        MVMint32 scid;
-                        MVMint32 objid;
+                        MVMuint32 scid;
+                        MVMuint32 objid;
                         if (MVM_bytecode_find_static_lexical_scref(tc, static_frame->body.cu,
                                 static_frame, i, &scid, &objid)) {
                             MVMObject *resolved;
@@ -471,7 +474,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         }
         else if (static_frame->body.outer) {
             /* Auto-close, and cache it in the static frame. */
-            MVMROOT2(tc, static_frame, code_ref, {
+            MVMROOT3(tc, static_frame, code_ref, static_code, {
                 MVM_frame_force_to_heap(tc, tc->cur_frame);
                 outer = autoclose(tc, static_frame->body.outer);
                 MVM_ASSIGN_REF(tc, &(static_code->common.header),
@@ -645,6 +648,7 @@ MVMFrame * MVM_frame_move_to_heap(MVMThreadContext *tc, MVMFrame *frame) {
     MVMROOT4(tc, new_cur_frame, update_caller, cur_to_promote, result, {
         while (cur_to_promote) {
             /* Allocate a heap frame. */
+            /* frame is safe from the GC as we wouldn't be here if it wasn't on the stack */
             MVMFrame *promoted = MVM_gc_allocate_frame(tc);
 
             /* Bump heap promotion counter, to encourage allocating this kind
@@ -749,9 +753,10 @@ MVMFrame * MVM_frame_debugserver_move_to_heap(MVMThreadContext *tc, MVMThreadCon
     MVMFrame *update_caller = NULL;
     MVMFrame *result = NULL;
     MVM_CHECK_CALLER_CHAIN(tc, cur_to_promote);
-    MVMROOT3(tc, new_cur_frame, update_caller, result, {
+    MVMROOT4(tc, new_cur_frame, update_caller, cur_to_promote, result, {
         while (cur_to_promote) {
             /* Allocate a heap frame. */
+            /* frame is safe from the GC as we wouldn't be here if it wasn't on the stack */
             MVMFrame *promoted = MVM_gc_allocate_frame(tc);
 
             /* Bump heap promotion counter, to encourage allocating this kind
@@ -913,7 +918,7 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
     }
 
     /* Switch back to the caller frame if there is one. */
-    if (caller && returner != tc->thread_entry_frame) {
+    if (caller && (returner != tc->thread_entry_frame || tc->nested_interpreter)) {
 
        if (tc->jit_return_address != NULL) {
             /* on a JIT frame, exit to interpreter afterwards */
@@ -947,10 +952,27 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
                     su(tc, srd);
                 else if (!unwind && sr)
                     sr(tc, srd);
+                /* The special_return or special_unwind handler may schedule a
+                   finalizer call for the current runloop. If we are already in
+                   the top most call frame of the runloop, we need to replace
+                   the thread_entry_frame with the finalizer call. Otherwise we
+                   won't run the special code for ending the runloop when the
+                   finalizer returns. */
+                if (returner == tc->thread_entry_frame && tc->cur_frame != caller)
+                    tc->thread_entry_frame = tc->cur_frame;
             }
         }
 
-        return 1;
+        if (returner == tc->thread_entry_frame && tc->cur_frame == caller) {
+                tc->cur_frame = NULL;
+                return 0;
+        }
+        else {
+            /* We're either somewhere in a nested call or already up in the top
+               most frame but still need to run some finalizer, so keep the
+               runloop running */
+            return 1;
+        }
     }
     else {
         tc->cur_frame = NULL;
@@ -1067,6 +1089,9 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
             MVMObject    *handler;
             MVMCallsite *two_args_callsite;
 
+            if (return_value)
+                MVM_exception_throw_adhoc(tc, "return_value + exit_handler case NYI");
+
             /* Force the frame onto the heap, since we'll reference it from the
              * unwind data. */
             MVMROOT3(tc, frame, cur_frame, return_value, {
@@ -1091,8 +1116,6 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
                 ud->abs_addr = abs_addr;
                 ud->rel_addr = rel_addr;
                 ud->jit_return_label = jit_return_label;
-                if (return_value)
-                    MVM_exception_throw_adhoc(tc, "return_value + exit_handler case NYI");
                 MVM_frame_special_return(tc, cur_frame, continue_unwind, NULL, ud,
                     mark_unwind_data);
             }
@@ -1233,7 +1256,7 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
         effective_sf  = f->static_info;
     }
     else if (f->spesh_cand) {
-        MVMint32 i;
+        MVMuint32 i;
         flags = NULL;
         for (i = 0; i < f->spesh_cand->num_inlines; i++) {
             MVMStaticFrame *isf = f->spesh_cand->inlines[i].sf;
@@ -1251,7 +1274,7 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
     }
     flag  = flags ? flags[effective_idx] : -1;
     if (flag != -1 && static_env[effective_idx].o == NULL) {
-        MVMint32 scid, objid;
+        MVMuint32 scid, objid;
         if (MVM_bytecode_find_static_lexical_scref(tc, effective_sf->body.cu,
                 effective_sf, effective_idx, &scid, &objid)) {
             MVMSerializationContext *sc;
@@ -1266,7 +1289,7 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
-            MVMROOT(tc, f, {
+            MVMROOT2(tc, f, effective_sf, {
                 resolved = MVM_sc_get_object(tc, sc, objid);
             });
             MVM_ASSIGN_REF(tc, &(effective_sf->common.header),
@@ -1301,11 +1324,8 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
 MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
             if (entry) {
                 if (MVM_LIKELY(cur_frame->static_info->body.lexical_types[entry->value] == type)) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1338,10 +1358,8 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
 MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister value) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
                     if (type == MVM_reg_obj || type == MVM_reg_str) {
@@ -1374,7 +1392,10 @@ MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *
 
 /* Finds a lexical in the outer frame, throwing if it's not there. */
 MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name) {
-    MVMRegister *r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
+    MVMRegister *r;
+    MVMROOT(tc, name, {
+        r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
+    });
     if (MVM_LIKELY(r != NULL))
         return r->o;
     else {
@@ -1389,11 +1410,8 @@ MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString
  * the specified frame. Only works if it's an object lexical.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
-        if (lexical_names) {
-            /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
-            MVMLexicalRegistry *entry;
-            MVM_HASH_GET(tc, lexical_names, name, entry)
+        if (cur_frame->static_info->body.num_lexicals) {
+            MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1419,12 +1437,14 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString
  * that is done depends on the frame walker setup. */
 MVMRegister * MVM_frame_lexical_lookup_using_frame_walker(MVMThreadContext *tc,
         MVMSpeshFrameWalker *fw, MVMString *name) {
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&(name));
     while (MVM_spesh_frame_walker_next(tc, fw)) {
         MVMRegister *found;
         MVMuint16 found_kind;
         if (MVM_spesh_frame_walker_get_lex(tc, fw, name, &found, &found_kind, 1, NULL)) {
             MVM_spesh_frame_walker_cleanup(tc, fw);
             if (found_kind == MVM_reg_obj) {
+                MVM_gc_root_temp_pop(tc);
                 return found;
             }
             else {
@@ -1437,6 +1457,7 @@ MVMRegister * MVM_frame_lexical_lookup_using_frame_walker(MVMThreadContext *tc,
         }
     }
     MVM_spesh_frame_walker_cleanup(tc, fw);
+    MVM_gc_root_temp_pop(tc);
     return NULL;
 }
 
@@ -1505,6 +1526,7 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
     /* Traverse with the frame walker. */
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&initial_frame);
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&last_real_frame);
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&name);
     while (MVM_spesh_frame_walker_next(tc, fw)) {
         MVMRegister *result;
 
@@ -1529,7 +1551,7 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
                         tc->instance->dynvar_log_lasttime = uv_hrtime();
                     }
                     *found_frame = last_real_frame;
-                    MVM_gc_root_temp_pop_n(tc, 2);
+                    MVM_gc_root_temp_pop_n(tc, 3);
                     MVM_spesh_frame_walker_cleanup(tc, fw);
                     return result;
                 }
@@ -1557,14 +1579,14 @@ MVMRegister * MVM_frame_find_dynamic_using_frame_walker(MVMThreadContext *tc,
                 MVM_free(c_name);
                 tc->instance->dynvar_log_lasttime = uv_hrtime();
             }
-            MVM_gc_root_temp_pop_n(tc, 2);
+            MVM_gc_root_temp_pop_n(tc, 3);
             MVM_spesh_frame_walker_cleanup(tc, fw);
             return result;
         }
     }
 
     /* Not found. */
-    MVM_gc_root_temp_pop_n(tc, 2);
+    MVM_gc_root_temp_pop_n(tc, 3);
     MVM_spesh_frame_walker_cleanup(tc, fw);
     if (dlog) {
         fprintf(dlog, "N %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
@@ -1646,7 +1668,10 @@ MVMObject * MVM_frame_getdynlex(MVMThreadContext *tc, MVMString *name, MVMFrame
 void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *value, MVMFrame *cur_frame) {
     MVMuint16 type;
     MVMFrame *found_frame;
-    MVMRegister *lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 0, &found_frame);
+    MVMRegister *lex_reg;
+    MVMROOT2(tc, name, value, {
+        lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 0, &found_frame);
+    });
     if (!lex_reg) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
@@ -1677,10 +1702,8 @@ void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *valu
 /* Returns the storage unit for the lexical in the specified frame. Does not
  * try to vivify anything - gets exactly what is there. */
 MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *name) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (MVM_LIKELY(lexical_names != NULL)) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
+    if (MVM_LIKELY(f->static_info->body.num_lexicals != 0)) {
+        MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, f->static_info, name);
         if (entry)
             return &f->env[entry->value];
     }
@@ -1694,10 +1717,8 @@ MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *na
 
 /* Returns the storage unit for the lexical in the specified frame. */
 MVMRegister * MVM_frame_try_get_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *name, MVMuint16 type) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (lexical_names) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
+    if (f->static_info->body.num_lexicals) {
+        MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, f->static_info, name);
         if (entry && f->static_info->body.lexical_types[entry->value] == type) {
             MVMRegister *result = &f->env[entry->value];
             if (type == MVM_reg_obj && !result->o)
@@ -1742,10 +1763,8 @@ MVMuint16 MVM_frame_translate_to_primspec(MVMThreadContext *tc, MVMuint16 kind)
 
 /* Returns the primitive type specification for a lexical. */
 MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMString *name) {
-    MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (lexical_names) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
+    if (f->static_info->body.num_lexicals) {
+        MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, f->static_info, name);
         if (entry)
             return MVM_frame_translate_to_primspec(tc,
                     f->static_info->body.lexical_types[entry->value]);
