@@ -71,6 +71,7 @@ void MVM_load_bytecode_buffer_to_cu(MVMThreadContext *tc, MVMObject *buf, MVMReg
     memcpy(data_start, (MVMuint8 *)(((MVMArray *)buf)->body.slots.i8 + ((MVMArray *)buf)->body.start), data_size);
 
     cu = MVM_cu_from_bytes(tc, data_start, data_size);
+    cu->body.deallocate = MVM_DEALLOCATE_FREE;
     res->o = (MVMObject *)cu;
 
     if (cu->body.deserialize_frame) {
@@ -101,7 +102,7 @@ void MVM_load_bytecode(MVMThreadContext *tc, MVMString *filename) {
     }
 
     /* Otherwise, load from disk. */
-    MVMROOT2(tc, cu, filename, {
+    MVMROOT(tc, filename, {
         char *c_filename = MVM_string_utf8_c8_encode_C_string(tc, filename);
         /* XXX any exception from MVM_cu_map_from_file needs to be handled
          *     and c_filename needs to be freed */
@@ -113,8 +114,10 @@ void MVM_load_bytecode(MVMThreadContext *tc, MVMString *filename) {
         run_comp_unit(tc, cu);
 
         loaded_name = MVM_calloc(1, sizeof(MVMLoadedCompUnitName));
+        MVM_HASH_BIND_FREE(tc, tc->instance->loaded_compunits, filename, loaded_name, {
+            MVM_free(loaded_name);
+        });
         loaded_name->filename = filename;
-        MVM_HASH_BIND(tc, tc->instance->loaded_compunits, filename, loaded_name);
     });
 
 LEAVE:
@@ -127,7 +130,7 @@ void MVM_load_bytecode_fh(MVMThreadContext *tc, MVMObject *oshandle, MVMString *
     if (REPR(oshandle)->ID != MVM_REPR_ID_MVMOSHandle)
         MVM_exception_throw_adhoc(tc, "loadbytecodefh requires an object with REPR MVMOSHandle");
 
-    MVMROOT2(tc, cu, filename, {
+    MVMROOT(tc, filename, {
         MVMuint64 pos = MVM_io_tell(tc, oshandle);
         cu = MVM_cu_map_from_file_handle(tc, MVM_io_fileno(tc, oshandle), pos);
         cu->body.filename = filename;
