@@ -317,7 +317,8 @@ char * MVM_nativecall_unmarshal_string(MVMThreadContext *tc, MVMObject *value, M
     }
 }
 
-static void unmarshal_error(MVMThreadContext *tc, char *desired_repr, MVMObject *value, MVMint16 unmarshal_kind) {
+MVM_NO_RETURN static void unmarshal_error(MVMThreadContext *tc, char *desired_repr, MVMObject *value, MVMint16 unmarshal_kind) MVM_NO_RETURN_ATTRIBUTE;
+MVM_NO_RETURN static void unmarshal_error(MVMThreadContext *tc, char *desired_repr, MVMObject *value, MVMint16 unmarshal_kind) {
     if (unmarshal_kind == MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC) {
         MVM_exception_throw_adhoc(tc,
             "NativeCall conversion expected type with %s representation, but got a %s (%s)", desired_repr, REPR(value)->name, MVM_6model_get_debug_name(tc, value));
@@ -573,6 +574,7 @@ MVMJitGraph *MVM_nativecall_jit_graph_for_caller_code(
         || body->ret_type == MVM_NATIVECALL_ARG_LONGLONG
         || body->ret_type == MVM_NATIVECALL_ARG_ULONGLONG
     ) {
+        call_node->u.call.rv_type = body->ret_type;
         init_box_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_int, restype, dst);
     }
     else if (body->ret_type == MVM_NATIVECALL_ARG_CPOINTER) {
@@ -635,28 +637,27 @@ MVMJitCode *create_caller_code(MVMThreadContext *tc, MVMNativeCallBody *body) {
 
 void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigned int interval_id) {
     /* Try to load the library. */
-    body->lib_handle = MVM_nativecall_load_lib(body->lib_name[0] ? body->lib_name : NULL);
+    DLLib *lib_handle = MVM_nativecall_load_lib(body->lib_name[0] ? body->lib_name : NULL);
 
-    if (!body->lib_handle) {
+    if (!lib_handle) {
         char *waste[] = { body->lib_name, NULL };
-        MVM_free(body->sym_name);
+        MVM_free_null(body->sym_name);
         body->lib_name = NULL;
-        body->sym_name = NULL;
         if (interval_id)
             MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
-        MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate native library '%s': %s", body->lib_name, dlerror());
+        MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate native library '%s': %s", waste[0], dlerror());
     }
 
     if (!body->entry_point) {
-        body->entry_point = MVM_nativecall_find_sym(body->lib_handle, body->sym_name);
+        body->entry_point = MVM_nativecall_find_sym(lib_handle, body->sym_name);
         if (!body->entry_point) {
             char *waste[] = { body->sym_name, body->lib_name, NULL };
-            body->lib_name = NULL;
             body->sym_name = NULL;
+            body->lib_name = NULL;
             if (interval_id)
                 MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
             MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate symbol '%s' in native library '%s'",
-                body->sym_name, body->lib_name);
+                waste[0], waste[1]);
         }
     }
 
@@ -665,6 +666,10 @@ void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigne
     }
     else
         body->jitcode = NULL;
+
+    /* Initialize body->lib_handle as late as possible, so get_int won't report
+       that the NativeCall site is set up too early */
+    body->lib_handle = lib_handle;
 }
 
 /* Builds up a native call site out of the supplied arguments. */
@@ -679,11 +684,17 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
 
     MVMObject *entry_point_o = (MVMObject *)MVM_repr_at_key_o(tc, ret_info,
         tc->instance->str_consts.entry_point);
+    MVMObject *resolve_lib_name = (MVMObject *)MVM_repr_at_key_o(tc, ret_info,
+        tc->instance->str_consts.resolve_lib_name);
+    MVMObject *resolve_lib_name_arg = (MVMObject *)MVM_repr_at_key_o(tc, ret_info,
+        tc->instance->str_consts.resolve_lib_name_arg);
 
     /* Initialize the object; grab native call part of its body. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
 
     body->lib_name = lib_name;
+    MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name, resolve_lib_name);
+    MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name_arg, resolve_lib_name_arg);
 
     /* Try to locate the symbol. */
     if (entry_point_o && !MVM_is_null(tc, entry_point_o)) {
@@ -1174,9 +1185,53 @@ MVMThreadContext * MVM_nativecall_find_thread_context(MVMInstance *instance) {
     return tc;
 }
 
+typedef struct ResolverData {
+    MVMObject *site;
+    MVMRegister args[1];
+} ResolverData;
+static void callback_invoke(MVMThreadContext *tc, void *data) {
+    /* Invoke the coderef, to set up the nested interpreter. */
+    ResolverData *r = (ResolverData*)data;
+    MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, r->site);
+    MVMObject *code = body->resolve_lib_name;
+    MVMCallsite *callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_INV_ARG);
+    r->args[0].o = body->resolve_lib_name_arg;
+    STABLE(code)->invoke(tc, code, callsite, r->args);
+
+    /* Ensure we exit interp after callback. */
+    tc->thread_entry_frame = tc->cur_frame;
+}
+void MVM_nativecall_restore_library(MVMThreadContext *tc, MVMNativeCallBody *body, MVMObject *root) {
+    if (!MVM_is_null(tc, body->resolve_lib_name) && !MVM_is_null(tc, body->resolve_lib_name_arg)) {
+        MVMRegister res = {NULL};
+        ResolverData data = {root, {NULL}};
+
+        MVM_interp_run_nested(tc, callback_invoke, &data, &res);
+
+        /* Handle return value. */
+        if (res.o) {
+            MVMContainerSpec const *contspec = STABLE(res.o)->container_spec;
+            if (contspec && contspec->fetch_never_invokes)
+                contspec->fetch(tc, res.o, &res);
+        }
+
+        body->lib_name = MVM_string_utf8_encode_C_string(tc, MVM_repr_get_str(tc, res.o));
+    }
+    if (body->lib_name && body->sym_name && !body->lib_handle) {
+        MVM_nativecall_setup(tc, body, 0);
+    }
+}
+
 void MVM_nativecall_invoke_jit(MVMThreadContext *tc, MVMObject *site) {
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
-    MVMJitCode * const jitcode = body->jitcode;
 
+    if (MVM_UNLIKELY(!body->lib_handle)) {
+        MVMROOT(tc, site, {
+            MVM_nativecall_restore_library(tc, body, site);
+        });
+        body = MVM_nativecall_get_nc_body(tc, site);
+    }
+
+    MVMJitCode * const jitcode = body->jitcode;
     jitcode->func_ptr(tc, *tc->interp_cu, jitcode->labels[0]);
 }
