@@ -87,9 +87,10 @@ static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMO
 
     if (!callback_data_head) {
         callback_data_head = MVM_malloc(sizeof(MVMNativeCallbackCacheHead));
+        MVM_HASH_BIND_FREE(tc, tc->native_callback_cache, cuid, callback_data_head, {
+            MVM_free(callback_data_head);
+        });
         callback_data_head->head = NULL;
-
-        MVM_HASH_BIND(tc, tc->native_callback_cache, cuid, callback_data_head);
     }
 
     callback_data_handle = &(callback_data_head->head);
@@ -689,6 +690,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     handle_ret(tc, unsigned long long, ffi_arg, MVM_nativecall_make_int);
                     break;
                 default:
+                    MVM_gc_mark_thread_unblocked(tc);
                     MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
             }
