@@ -277,13 +277,6 @@ struct MVMThreadContext {
      * like I/O, which grab a mutex but may throw an exception. */
     uv_mutex_t *ex_release_mutex;
 
-    /* Memory buffer pointing to the last thing we serialized, intended to go
-     * into the next compilation unit we write. Also the serialized string
-     * heap, which will be used to seed the compilation unit string heap. */
-    MVMint32      serialized_size;
-    char         *serialized;
-    MVMObject    *serialized_string_heap;
-
     /* Serialization context write barrier disabled depth (anything non-zero
      * means disabled). */
     MVMint32           sc_wb_disable_depth;
@@ -297,6 +290,10 @@ struct MVMThreadContext {
     MVMObject     *cur_dispatcher;
     MVMObject     *cur_dispatcher_for;
 
+    /* Dispatcher to return control to when current dispatcher exhausts. */
+    MVMObject     *next_dispatcher;
+    MVMObject     *next_dispatcher_for;
+
     /* Cache of native code callback data. */
     MVMNativeCallbackCacheHead *native_callback_cache;
 
@@ -333,6 +330,8 @@ struct MVMThreadContext {
 
     MVMuint32 cur_file_idx;
     MVMuint32 cur_line_no;
+
+    int nested_interpreter;
 };
 
 MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance);
