@@ -84,7 +84,7 @@ static void start_thread(void *data) {
     MVM_debugserver_notify_thread_creation(tc);
 
     /* Enter the interpreter, to run code. */
-    MVM_interp_run(tc, thread_initial_invoke, ts);
+    MVM_interp_run(tc, thread_initial_invoke, ts, NULL);
 
     MVM_debugserver_notify_thread_destruction(tc);
 
@@ -122,7 +122,9 @@ void MVM_thread_run(MVMThreadContext *tc, MVMObject *thread_obj) {
             MVM_profile_log_thread_created(tc, child_tc);
         }
 
-        /* Mark thread as GC blocked until the thread actually starts. */
+        /* Mark thread as GC blocked until the thread actually starts.
+         * Contrary to other callsites, MVM_gc_mark_thread_blocked cannot enter
+         * the GC here, as child_tc is not yet in the instance's threads list. */
         MVM_gc_mark_thread_blocked(child_tc);
 
         /* Create thread state, to pass to the thread start callback. */
@@ -266,7 +268,8 @@ MVMint32 MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head)
             case MVM_thread_stage_starting:
             case MVM_thread_stage_waiting:
             case MVM_thread_stage_started:
-                alive++; /* fall through intended */
+                alive++;
+                MVM_FALLTHROUGH
             case MVM_thread_stage_exited:
             case MVM_thread_stage_clearing_nursery:
                 /* push it to the new starting list */
