@@ -5,7 +5,7 @@
 /* If we have the job of doing GC for a thread, we add it to our work
  * list. */
 static void add_work(MVMThreadContext *tc, MVMThreadContext *stolen) {
-    MVMint32 i;
+    MVMuint32 i;
     for (i = 0; i < tc->gc_work_count; i++)
         if (tc->gc_work[i].tc == stolen)
             return;
@@ -27,7 +27,7 @@ static void add_work(MVMThreadContext *tc, MVMThreadContext *stolen) {
 static MVMuint32 signal_one_thread(MVMThreadContext *tc, MVMThreadContext *to_signal) {
     /* Loop here since we may not succeed first time (e.g. the status of the
      * thread may change between the two ways we try to twiddle it). */
-    int had_suspend_request = 0;
+    unsigned int had_suspend_request = 0;
     while (1) {
         AO_t current = MVM_load(&to_signal->gc_status);
         switch (current) {
@@ -407,7 +407,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
 
     MVMuint8 is_coordinator;
 
-    MVMuint64 start_time;
+    MVMuint64 start_time = 0;
 
     unsigned int interval_id;
 
@@ -466,7 +466,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
 
         data[0] = MVM_load(&tc->instance->gc_seq_number);
         data[1] = start_time / 1000;
-        data[2] = (start_time - tc->instance->subscriptions.vm_startup_time) / 1000;
+        data[2] = (start_time - tc->instance->subscriptions.vm_startup_hrtime) / 1000;
         data[3] = (end_time - start_time) / 1000;
         data[4] = gen == MVMGCGenerations_Both;
         data[5] = tc->gc_promoted_bytes;
@@ -499,7 +499,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
 
     /* Try to start the GC run. */
     if (MVM_trycas(&tc->instance->gc_start, 0, 1)) {
-        MVMThread *last_starter = NULL;
         MVMuint32 num_threads = 0;
 
         /* Stash us as the thread to blame for this GC run (used to give it a
@@ -626,8 +625,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
  * MVMSUSPENDSTATUS_MASK.
  *   */
 void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
-    AO_t curr;
-
     GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Entered from interrupt\n");
 
 
