@@ -86,10 +86,12 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Set up instance data structure. */
     instance = MVM_calloc(1, sizeof(MVMInstance));
 
-    instance->subscriptions.vm_startup_time = uv_hrtime();
-
     /* Create the main thread's ThreadContext and stash it. */
     instance->main_thread = MVM_tc_create(NULL, instance);
+
+    instance->subscriptions.vm_startup_hrtime = uv_hrtime();
+    instance->subscriptions.vm_startup_now = MVM_proc_time_n(instance->main_thread);
+
 #if MVM_HASH_RANDOMIZE
     /* Get the 128-bit hashSecret */
     MVM_getrandom(instance->main_thread, instance->hashSecrets, sizeof(MVMuint64) * 2);
@@ -435,7 +437,7 @@ static void run_deserialization_frame(MVMThreadContext *tc, MVMCompUnit *cu) {
     if (cu->body.deserialize_frame) {
         MVMint8 spesh_enabled_orig = tc->instance->spesh_enabled;
         tc->instance->spesh_enabled = 0;
-        MVM_interp_run(tc, toplevel_initial_invoke, cu->body.deserialize_frame);
+        MVM_interp_run(tc, toplevel_initial_invoke, cu->body.deserialize_frame, NULL);
         tc->instance->spesh_enabled = spesh_enabled_orig;
     }
 }
@@ -446,19 +448,17 @@ void MVM_vm_run_file(MVMInstance *instance, const char *filename) {
     MVMThreadContext *tc = instance->main_thread;
     MVMCompUnit      *cu = MVM_cu_map_from_file(tc, filename);
 
-    MVMROOT(tc, cu, {
-        /* The call to MVM_string_utf8_decode() may allocate, invalidating the
-           location cu->body.filename */
-        MVMString *const str = MVM_string_utf8_c8_decode(tc, instance->VMString, filename, strlen(filename));
-        cu->body.filename = str;
-        MVM_gc_write_barrier_hit(tc, (MVMCollectable *)cu);
+    /* The call to MVM_string_utf8_decode() may allocate, invalidating the
+       location cu->body.filename */
+    MVMString *const str = MVM_string_utf8_c8_decode(tc, instance->VMString, filename, strlen(filename));
+    cu->body.filename = str;
+    MVM_gc_write_barrier_hit(tc, (MVMCollectable *)cu);
 
-        /* Run the deserialization frame, if any. */
-        run_deserialization_frame(tc, cu);
-    });
+    /* Run the deserialization frame, if any. */
+    run_deserialization_frame(tc, cu);
 
     /* Run the entry-point frame. */
-    MVM_interp_run(tc, toplevel_initial_invoke, cu->body.main_frame);
+    MVM_interp_run(tc, toplevel_initial_invoke, cu->body.main_frame, NULL);
 }
 
 /* Loads bytecode from memory and runs it. */
@@ -468,10 +468,10 @@ void MVM_vm_run_bytecode(MVMInstance *instance, MVMuint8 *bytes, MVMuint32 size)
     MVMCompUnit      *cu = MVM_cu_from_bytes(tc, bytes, size);
 
     /* Run the deserialization frame, if any. */
-    MVMROOT(tc, cu, { run_deserialization_frame(tc, cu); });
+    run_deserialization_frame(tc, cu);
 
     /* Run the entry-point frame. */
-    MVM_interp_run(tc, toplevel_initial_invoke, cu->body.main_frame);
+    MVM_interp_run(tc, toplevel_initial_invoke, cu->body.main_frame, NULL);
 }
 
 /* Loads bytecode from the specified file name and dumps it. */
@@ -481,7 +481,7 @@ void MVM_vm_dump_file(MVMInstance *instance, const char *filename) {
     MVMCompUnit      *cu = MVM_cu_map_from_file(tc, filename);
     char *dump = MVM_bytecode_dump(tc, cu);
     size_t dumplen = strlen(dump);
-    int position = 0;
+    size_t position = 0;
 
     /* libuv already set up stdout to be nonblocking, but it can very well be
      * we encounter EAGAIN (Resource temporarily unavailable), so we need to
@@ -560,6 +560,9 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     MVM_spesh_worker_join(instance->main_thread);
     MVM_io_eventloop_destroy(instance->main_thread);
 
+    /* Run the normal GC one more time to actually collect the spesh thread */
+    MVM_gc_enter_from_allocator(instance->main_thread);
+
     /* Run the GC global destruction phase. After this,
      * no 6model object pointers should be accessed. */
     MVM_gc_global_destruction(instance->main_thread);
@@ -694,13 +697,14 @@ void MVM_vm_set_prog_name(MVMInstance *instance, const char *prog_name) {
 void MVM_vm_event_subscription_configure(MVMThreadContext *tc, MVMObject *queue, MVMObject *config) {
     MVMString *gcevent;
     MVMString *speshoverviewevent;
+    MVMString *startup_time;
 
     MVMROOT2(tc, queue, config, {
         if (!IS_CONCRETE(config)) {
             MVM_exception_throw_adhoc(tc, "vmeventsubscribe requires a concrete configuration hash (got a %s type object)", MVM_6model_get_debug_name(tc, config));
         }
 
-        if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue && !MVM_is_null(tc, queue) || !IS_CONCRETE(queue)) {
+        if ((REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue && !MVM_is_null(tc, queue)) || !IS_CONCRETE(queue)) {
             MVM_exception_throw_adhoc(tc, "vmeventsubscribe requires a concrete ConcBlockingQueue (got a %s)", MVM_6model_get_debug_name(tc, queue));
         }
 
@@ -713,6 +717,9 @@ void MVM_vm_event_subscription_configure(MVMThreadContext *tc, MVMObject *queue,
         gcevent = MVM_string_utf8_decode(tc, tc->instance->VMString, "gcevent", 7);
         MVMROOT(tc, gcevent, {
             speshoverviewevent = MVM_string_utf8_decode(tc, tc->instance->VMString, "speshoverviewevent", 18);
+            MVMROOT(tc, speshoverviewevent, {
+                startup_time = MVM_string_utf8_decode(tc, tc->instance->VMString, "startup_time", 12);
+            });
         });
 
         if (MVM_repr_exists_key(tc, config, gcevent)) {
@@ -744,6 +751,20 @@ void MVM_vm_event_subscription_configure(MVMThreadContext *tc, MVMObject *queue,
                 MVM_exception_throw_adhoc(tc, "vmeventsubscribe expects value at 'speshoverviewevent' key to be null (to unsubscribe) or a VMArray of int64 type object, got a %s%s%s (%s)", IS_CONCRETE(value) ? "concrete " : "", MVM_6model_get_debug_name(tc, value), IS_CONCRETE(value) ? "" : " type object", REPR(value)->name);
             }
         }
+
+        if (MVM_repr_exists_key(tc, config, startup_time)) {
+            /* Value is ignored, it will just be overwritten. */
+            MVMObject *value = NULL; 
+            MVMROOT3(tc, gcevent, speshoverviewevent, startup_time, {
+                    value = MVM_repr_box_num(tc, tc->instance->boot_types.BOOTNum, tc->instance->subscriptions.vm_startup_now);
+            });
+
+            if (MVM_is_null(tc, value)) {
+                uv_mutex_unlock(&tc->instance->subscriptions.mutex_event_subscription);
+                MVM_exception_throw_adhoc(tc, "vmeventsubscribe was unable to create a Num object to hold the vm startup time.");
+            }
+            MVM_repr_bind_key_o(tc, config, startup_time, value);
+        }
     });
 
     uv_mutex_unlock(&tc->instance->subscriptions.mutex_event_subscription);
