@@ -51,6 +51,7 @@ void MVM_profile_heap_start(MVMThreadContext *tc, MVMObject *config) {
         MVM_repr_at_key_o(tc, config, tc->instance->str_consts.path));
 
     if (MVM_is_null(tc, (MVMObject*)path_str)) {
+        MVM_free(col);
         MVM_exception_throw_adhoc(tc, "Didn't specify a path for the heap snapshot profiler");
     }
 
@@ -59,6 +60,7 @@ void MVM_profile_heap_start(MVMThreadContext *tc, MVMObject *config) {
     col->fh = MVM_platform_fopen(path, "w");
 
     if (!col->fh) {
+        MVM_free(col);
         char *waste[2] = {path, NULL};
         MVM_exception_throw_adhoc_free(tc, waste, "Couldn't open heap snapshot target file %s: %s", path, strerror(errno));
     }
@@ -371,9 +373,20 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
     MVMCompUnit *cu = sf->body.cu;
     MVMBytecodeAnnotation *ann = MVM_bytecode_resolve_annotation(tc, &(sf->body), 0);
     MVMuint64 line = ann ? ann->line_number : 1;
-    MVMuint64 file_idx = ann && ann->filename_string_heap_index < cu->body.num_strings
-        ? get_vm_string_index(tc, ss, MVM_cu_string(tc, cu, ann->filename_string_heap_index))
-        : get_vm_string_index(tc, ss, cu->body.filename);
+    MVMString *file_name = ann && ann->filename_string_heap_index < cu->body.num_strings
+        ? MVM_cu_string(tc, cu, ann->filename_string_heap_index)
+        : cu->body.filename;
+
+    /* We're running as part of gc_finalize. By now, live objects have been marked
+       but not yet collected. If the file name was just deserialized from the string
+       heap by MVM_cu_string, it will have been allocated in gen2 directly, but not
+       marked as live for this gc run. Do it here to prevent it from getting freed
+       after taking this heap snapshot, while it's actually still referenced from the
+       comp unit's strings list */
+    if (file_name && file_name->common.header.flags & MVM_CF_SECOND_GEN)
+        file_name->common.header.flags |= MVM_CF_GEN2_LIVE;
+
+    MVMuint64 file_idx = get_vm_string_index(tc, ss, file_name);
 
     MVMuint64 i;
     MVMHeapSnapshotStaticFrame *s;
@@ -799,9 +812,7 @@ static void destroy_current_heap_snapshot(MVMThreadContext *tc) {
 
     MVM_free(col->snapshot->collectables);
     MVM_free(col->snapshot->references);
-    MVM_free(col->snapshot);
-
-    col->snapshot = NULL;
+    MVM_free_null(col->snapshot);
 }
 
 /* Frees all memory associated with the heap snapshot. */
@@ -868,7 +879,6 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
 
     size_t outSize = ZSTD_CStreamOutSize();
 
-    ZSTD_inBuffer inbuf;
     ZSTD_outBuffer outbuf;
 
     size_t size_position = ftell(fh);
@@ -879,7 +889,7 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
     MVMuint16 elem_size_to_write = elem_size;
 
     size_t written_total = 0;
-    size_t original_total = count * elem_size;
+    /*size_t original_total = count * elem_size;*/
 
     char *out_buffer = MVM_malloc(outSize);
     outbuf.dst = out_buffer;
@@ -892,8 +902,8 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
     }
 
     {
-        char namebuf[8];
-        strncpy(namebuf, name, 8);
+        char namebuf[8] = {0};
+        memcpy(namebuf, name, 8);
         fwrite(namebuf, 8, 1, fh);
     }
 
@@ -966,7 +976,6 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
 void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollection *col) {
     FILE *fh = col->fh;
     MVMuint64 i = col->strings_written;
-    MVMHeapDumpIndex *index = col->index;
 
     size_t return_value = 0;
     size_t result_buffer_size = 2048;
@@ -975,8 +984,6 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
 
     ZSTD_CStream *cstream;
 
-    size_t outSize = ZSTD_CStreamOutSize();
-
     ZSTD_inBuffer inbuf;
     ZSTD_outBuffer outbuf;
 
@@ -1290,12 +1297,12 @@ static void make_leaderboards(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
 
     leaderboard boards[4][LEADERBOARDS_TOP_SPOTS];
 
-    char *descriptions[4] = {
+    /*char *descriptions[4] = {
         "types_by_count",
         "frames_by_count",
         "types_by_size",
         "frames_by_size",
-    };
+    };*/
 
     if (!stats)
         return;
@@ -1320,7 +1327,7 @@ static void make_leaderboards(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
 
         /*fprintf(stderr, "  %s: [\n", descriptions[which]);*/
         for (i = 0; i < LEADERBOARDS_TOP_SPOTS; i++) {
-            char *name = NULL;
+            /*char *name = NULL;*/
             boards[which][i].tofi = data_body[i].identity;
             boards[which][i].value = data_body[i].value;
 
@@ -1346,11 +1353,10 @@ static void make_leaderboards(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
 }
 
 void snapshot_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollection *col) {
-    MVMuint64 i = col->snapshot_idx;
-    MVMHeapDumpIndexSnapshotEntry *entry;
+    MVMHeapDumpIndexSnapshotEntry *entry = NULL;
 
     MVMHeapDumpTableOfContents *outer_toc = col->toplevel_toc;
-    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(sizeof(MVMHeapDumpTableOfContents), 1);
+    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(1, sizeof(MVMHeapDumpTableOfContents));
 
     /*fprintf(stderr, "creating a new TOC at %p\n", inner_toc);*/
 
@@ -1386,11 +1392,11 @@ void static_frames_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotColle
     MVMuint64 i;
     MVMHeapDumpIndex *index = col->index;
     FILE *fh = col->fh;
-    gzFile names_fh, cuid_fh, line_fh, file_fh;
 
     fputs("fram", fh);
 
 #if DUMP_EVERYTHING_RAW
+    gzFile names_fh, cuid_fh, line_fh, file_fh;
     names_fh = open_coll_file(col, "names");
     cuid_fh = open_coll_file(col, "cuid");
     line_fh = open_coll_file(col, "line");
@@ -1452,7 +1458,6 @@ void string_heap_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollect
     for (i = col->strings_written; i < col->num_strings; i++) {
         char *str = col->strings[i];
         MVMuint64 output_size = strlen(str);
-        MVMuint32 smaller_size = strlen(str);
 
         /* Every string is stored as a 64 bit integer length followed by utf8-
          * encoded string data. */
@@ -1466,6 +1471,7 @@ void string_heap_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollect
 
 
 #if DUMP_EVERYTHING_RAW
+        MVMuint32 smaller_size = strlen(str);
         gzfwrite(&smaller_size, sizeof(MVMuint32), 1, str_fh);
         gzfwrite(str, sizeof(MVMuint8), smaller_size, str_fh);
 #endif
@@ -1495,7 +1501,6 @@ void references_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollecti
     MVMuint64 halfway;
     FILE *fh = col->fh;
     MVMHeapSnapshot *s = col->snapshot;
-    gzFile descr_fh, kind_fh, cindex_fh;
 
     fputs("refs", fh);
     fwrite(&s->num_references, sizeof(MVMuint64), 1, fh);
@@ -1507,6 +1512,7 @@ void references_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollecti
     halfway = s->num_references / 2 - 1;
 
 #if DUMP_EVERYTHING_RAW
+    gzFile descr_fh, kind_fh, cindex_fh;
     descr_fh = open_coll_file(col, "descrs");
     kind_fh = open_coll_file(col, "kinds");
     cindex_fh = open_coll_file(col, "cindex");
@@ -1554,7 +1560,7 @@ void references_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollecti
             entry->full_refs_size += sizeof(MVMuint16) * 2 + 2;
         }
         else {
-            MVMuint8 descr8, kind8, index8;
+            MVMuint8 kind8, index8;
             kind8  = kind;
             index8 = cindex;
             fputc('0', fh);
@@ -1589,11 +1595,11 @@ void types_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
     MVMuint64 i;
     MVMHeapDumpIndex *index = col->index;
     FILE *fh = col->fh;
-    gzFile repr_fh, type_fh;
 
     fputs("type", fh);
 
 #if DUMP_EVERYTHING_RAW
+    gzFile repr_fh, type_fh;
     repr_fh = open_coll_file(col, "repr");
     type_fh = open_coll_file(col, "type");
 #endif
@@ -1635,7 +1641,6 @@ void collectables_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollec
     MVMuint64 i;
     FILE *fh = col->fh;
     MVMHeapSnapshot *s = col->snapshot;
-    gzFile kind_fh, tofi_fh, collsize_fh, unmansize_fh, refstart_fh, refcount_fh;
 
     fputs("coll", fh);
 
@@ -1647,6 +1652,7 @@ void collectables_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollec
 
     
 #if DUMP_EVERYTHING_RAW
+    gzFile kind_fh, tofi_fh, collsize_fh, unmansize_fh, refstart_fh, refcount_fh;
     kind_fh = open_coll_file(col, "kind");
     tofi_fh = open_coll_file(col, "tofi");
     collsize_fh = open_coll_file(col, "size");
@@ -1834,7 +1840,7 @@ void finish_collection_to_filehandle(MVMThreadContext *tc, MVMHeapSnapshotCollec
 
 #if MVM_HEAPSNAPSHOT_FORMAT == 3
     {
-    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(sizeof(MVMHeapDumpTableOfContents), 1);
+    MVMHeapDumpTableOfContents *inner_toc = MVM_calloc(1, sizeof(MVMHeapDumpTableOfContents));
 
     /*fprintf(stderr, "finishing collection; creating a new TOC at %p\n", inner_toc);*/
 
