@@ -105,6 +105,7 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVMSpeshBB *bb = ig->entry;
     MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
             inliner->sf->body.cu->body.hll_config;
+    MVMuint32 seen_possible_deopt = 0;
     if (no_inline_info)
         *no_inline_info = NULL;
     while (bb) {
@@ -161,7 +162,9 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 }
             }
 
-            /* Check we don't have too many args for inlining to work out. */
+            /* Check we don't have too many args for inlining to work out,
+             * and that we're not reading args after a possible deopt
+             * instruction, because we can't uninline in such a case. */
             else if (opcode == MVM_OP_sp_getarg_o ||
                     opcode == MVM_OP_sp_getarg_i ||
                     opcode == MVM_OP_sp_getarg_n ||
@@ -170,6 +173,16 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     *no_inline_reason = "too many arguments to inline";
                     return 0;
                 }
+                if (seen_possible_deopt) {
+                    *no_inline_reason = "a deopt may happen before arguments are processed";
+                    return 0;
+                }
+            }
+            else if (opcode == MVM_OP_sp_paramnamesused) {
+                if (seen_possible_deopt) {
+                    *no_inline_reason = "a deopt may happen before arguments are processed";
+                    return 0;
+                }
             }
 
             /* Ext-ops need special care in inter-comp-unit inlines. */
@@ -180,6 +193,10 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     demand_extop(tc, target_cu, source_cu, ins->info);
             }
 
+            /* Record if this instruction would cause us to deopt. */
+            if (ins->info->may_cause_deopt)
+                seen_possible_deopt = 1;
+
             ins = ins->next;
         }
         bb = bb->linear_next;
@@ -223,7 +240,7 @@ static void add_deopt_usages(MVMThreadContext *tc, MVMSpeshGraph *g, MVMint32 *d
 
 /* Get the maximum inline size applicable to the specified static frame (it can
  * be configured by language). */
-int MVM_spesh_inline_get_max_size(MVMThreadContext *tc, MVMStaticFrame *sf) {
+MVMuint32 MVM_spesh_inline_get_max_size(MVMThreadContext *tc, MVMStaticFrame *sf) {
     return sf->body.cu->body.hll_config->max_inline_size;
 }
 
@@ -303,8 +320,10 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadConte
      * the args specialization). */
     ig = MVM_spesh_graph_create(tc, target_sf, 0, 1);
     MVM_spesh_args_from_callinfo(tc, ig, call_info, type_tuple);
-    MVM_spesh_facts_discover(tc, ig, NULL, 0);
-    MVM_spesh_optimize(tc, ig, NULL);
+    MVMROOT(tc, target_sf, {
+        MVM_spesh_facts_discover(tc, ig, NULL, 0);
+        MVM_spesh_optimize(tc, ig, NULL);
+    });
 
     /* See if it's inlineable; clean up if not. */
     if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
@@ -320,7 +339,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadConte
 }
 
 /* Finds the deopt index of the return. */
-static MVMint32 return_deopt_idx(MVMThreadContext *tc, MVMSpeshIns *invoke_ins) {
+static MVMuint32 return_deopt_idx(MVMThreadContext *tc, MVMSpeshIns *invoke_ins) {
     MVMSpeshAnn *ann = invoke_ins->annotations;
     while (ann) {
         if (ann->type == MVM_SPESH_ANN_DEOPT_ALL_INS)
@@ -384,7 +403,7 @@ static void fix_wval(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVMint64     idx      = to_fix->info->opcode == MVM_OP_wval
         ? to_fix->operands[2].lit_i16
         : to_fix->operands[2].lit_i64;
-    if (dep >= 0 && dep < sourcecu->body.num_scs) {
+    if (dep >= 0 && (MVMuint16)dep < sourcecu->body.num_scs) {
         MVMSerializationContext *sc = MVM_sc_get_sc(tc, sourcecu, dep);
         if (sc) {
             MVMuint16 otherdep;
@@ -520,7 +539,7 @@ static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                  MVMCallsite *cs) {
     MVMSpeshFacts **merged_facts;
     MVMuint16      *merged_fact_counts;
-    MVMint32        i, j, orig_inlines, total_inlines, orig_deopt_addrs,
+    MVMuint32        i, j, orig_inlines, total_inlines, orig_deopt_addrs,
                     orig_deopt_pea_mat_infos, impl_deopt_idx;
     MVMuint32       total_handlers = inliner->num_handlers + inlinee->num_handlers + 1;
     MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
@@ -541,11 +560,11 @@ static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     impl_deopt_idx = return_deopt_idx(tc, invoke_ins);
     MVM_VECTOR_INIT(regs_for_deopt, 0);
     for (i = 0; i < inliner->sf->body.num_locals; i++) {
-        MVMint32 vers = inliner->fact_counts[i];
+        MVMuint16 vers = inliner->fact_counts[i];
         for (j = 0; j < vers; j++) {
             MVMSpeshDeoptUseEntry *due = inliner->facts[i][j].usage.deopt_users;
             while (due) {
-                if (due->deopt_idx == impl_deopt_idx) {
+                if ((MVMuint32)due->deopt_idx == impl_deopt_idx) {
                     MVMSpeshOperand o;
                     o.reg.orig = i;
                     o.reg.i = j;
@@ -630,6 +649,14 @@ static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     ins->operands[0].reg.orig += inliner->num_locals;
                 }
             }
+            else if (opcode == MVM_OP_takenextdispatcher) {
+                /* If the inlining code doesn't set a next dispatcher, takenextdispatcher
+                 * in the inlinee can be assumed to always return null. */
+                if (!inliner->sets_nextdispatcher && !inlinee->sets_nextdispatcher) {
+                    ins->info = MVM_op_get_op(MVM_OP_null);
+                    ins->operands[0].reg.orig += inliner->num_locals;
+                }
+            }
             else {
                 if (!same_comp_unit) {
                     if (ins->info->opcode == MVM_OP_const_s) {
